package org.jfree.date;

import java.text.DateFormatSymbols;
import java.text.SimpleDateFormat;

public enum Month {
    JANUARY(1,1,31),
    FEBRUARY(2,1,28),
    MARCH(3,1,31),
    APRIL(4,2,30),
    MAY(5,2,31),
    JUNE(6,2,30),
    JULY(7,3,31),
    AUGUST(8,3,31),
    SEPTEMBER(9,3,30),
    OCTOBER(10,4,31),
    NOVEMBER(11,4,30),
    DECEMBER(12,4,31);

    private static final DateFormatSymbols
    	DATE_FORMAT_SYMBOLS = new SimpleDateFormat().getDateFormatSymbols();
    
    private int monthCode, quarter; 

    /** The number of days in each month in non leap years. */
    private int lastDayOfMonth;
    
    private Month(int monthCode, int quarter, int lastDayOfMonth) {
    	this.monthCode = monthCode;
    	this.quarter = quarter;
    	this.lastDayOfMonth = lastDayOfMonth;
    }
	
    public int getMonthCode() {
		return monthCode;
	}
    
    /**
     * Returns a string representing the supplied month.
     * <P>
     * The string returned is the long form of the month name taken from the 
     * default locale.
     *
     * @return a string representing the supplied month.
     */
    public String asString() {
    	return asString(false);
    }
    
    /**
     * Returns a string representing the supplied month.
     * <P>
     * The string returned is the long or short form of the month name taken 
     * from the default locale.
     *
     * @param shortened  if <code>true</code> return the abbreviation of the 
     *                   month.
     *
     * @return a string representing the supplied month.
     */
    public String asString(final boolean shortened) {

        final String[] months;

        if (shortened) {
            months = DATE_FORMAT_SYMBOLS.getShortMonths();
        }
        else {
            months = DATE_FORMAT_SYMBOLS.getMonths();
        }

        return months[monthCode - 1];
    }
    
    /**
     * Returns the quarter for the specified month.
     *
     * @param code  the month code (1-12).
     *
     * @return the quarter that the month belongs to.
     */
    public int getQuarter() {
    	return quarter;
    }

    /**
     * Returns the number of the last day of the month, taking into account 
     * leap years.
     *
     * @param month  the month.
     * @param yyyy  the year (in the range 1900 to 9999).
     *
     * @return the number of the last day of the month.
     */
    public int lastDayOfMonth(final int yyyy) {

        final int result = lastDayOfMonth;
        if (this != FEBRUARY) {
            return result;
        }
        else if (SerialDate.isLeapYear(yyyy)) {
            return result + 1;
        }
        else {
            return result;
        }
    }
    
    /**
     * Converts a string to a month code.
     * <P>
     * This method will return one of the constants JANUARY, FEBRUARY, ..., 
     * DECEMBER that corresponds to the string.  If the string is not 
     * recognised, this method returns -1.
     *
     * @param s  the string to parse.
     *
     * @return <code>-1</code> if the string is not parseable, the month of the
     *         year otherwise.
     */
    public static Month stringToMonthCode(String s) {

        final String[] shortMonthNames = DATE_FORMAT_SYMBOLS.getShortMonths();
        final String[] monthNames = DATE_FORMAT_SYMBOLS.getMonths();

        int result = -1;
        s = s.trim();

        // first try parsing the string as an integer (1-12)...
        try {
            result = Integer.parseInt(s);
        } catch (NumberFormatException e) {
            // suppress
        }

        // now search through the month names...
        if ((result < 1) || (result > 12)) {
            for (int i = 0; i < monthNames.length; i++) {
                if (s.equals(shortMonthNames[i])) {
                    result = i + 1;
                    break;
                }
                if (s.equals(monthNames[i])) {
                    result = i + 1;
                    break;
                }
            }
        }

        return Month.values()[result-1];
    }
}
